﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace TemporalSourceCover
{
#if DEBUG
    class TemporalGraphReachabilityState
    {
        public readonly Dictionary<TemporalGraphVertex, HashSet<TemporalGraphVertex>> BySource;
        public readonly Dictionary<TemporalGraphVertex, HashSet<TemporalGraphVertex>> ByTarget;

        public TemporalGraphReachabilityState(TemporalGraphReachability reachability)
        {
            BySource = reachability.BySource.ToDictionary(pair => pair.Key, pair => new HashSet<TemporalGraphVertex>(pair.Value));
            ByTarget = reachability.ByTarget.ToDictionary(pair => pair.Key, pair => new HashSet<TemporalGraphVertex>(pair.Value));
        }

        public static bool operator ==(TemporalGraphReachabilityState a, TemporalGraphReachabilityState b)
        {
            return a.BySource.Count == b.BySource.Count
                && a.ByTarget.Count == b.ByTarget.Count
                && a.BySource.All(pair => b.BySource.TryGetValue(pair.Key, out var targets) && pair.Value.Count == targets.Count && pair.Value.All(t => targets.Contains(t)))
                && a.ByTarget.All(pair => b.ByTarget.TryGetValue(pair.Key, out var sources) && pair.Value.Count == sources.Count && pair.Value.All(t => sources.Contains(t)));
        }

        public static bool operator !=(TemporalGraphReachabilityState a, TemporalGraphReachabilityState b) => !(a == b);

        public override bool Equals(object? obj)
        {
            return obj is TemporalGraphReachabilityState state &&
                this == state;
        }
    }
#endif
    public sealed class TemporalGraphReachabilityTransaction : IDisposable
    {
        private enum TransactionState
        {
            Active,
            Committed,
            RolledBack,
        }

        private readonly TemporalGraphReachabilityTransaction? _outerTransaction;

        private readonly TemporalGraphReachability _reachability;
        private TransactionState _state = TransactionState.Active;

        private readonly Dictionary<TemporalGraphVertex, List<TemporalGraphVertex>> _removedBySource = new();
        private readonly Dictionary<TemporalGraphVertex, List<TemporalGraphVertex>> _removedByTarget = new();

#if DEBUG
        private readonly TemporalGraphReachabilityState _initialState;
#endif

        public TemporalGraphReachabilityTransaction(TemporalGraphReachability reachability)
        {
            _reachability = reachability;

            _outerTransaction = reachability.CurrentTransaction;
            _reachability.CurrentTransaction = this;

#if DEBUG
            _initialState = new TemporalGraphReachabilityState(reachability);
#endif
        }

        public void Dispose()
        {
            if (_state == TransactionState.Active)
                throw new InvalidOperationException();

            _reachability.CurrentTransaction = _outerTransaction;
        }

        public void Commit()
        {
            if (_state != TransactionState.Active)
                throw new InvalidOperationException();

            _state = TransactionState.Committed;
        }

        public void Rollback()
        {
            if (_state != TransactionState.Active)
                throw new InvalidOperationException();

            foreach(var (source, targets) in _removedBySource)
            {
                if(!_reachability.BySource.TryGetValue(source, out var curTargets))
                {
                    curTargets = new HashSet<TemporalGraphVertex>();
                    _reachability.BySource.Add(source, curTargets);
                }
                curTargets.AddRange(targets);
            }

            foreach (var (target, sources) in _removedByTarget)
            {
                if (!_reachability.ByTarget.TryGetValue(target, out var curSources))
                {
                    curSources = new HashSet<TemporalGraphVertex>();
                    _reachability.ByTarget.Add(target, curSources);
                }
                curSources.AddRange(sources);
            }


            _state = TransactionState.RolledBack;

#if DEBUG
            var curState = new TemporalGraphReachabilityState(_reachability);
            if (_initialState != curState)
                throw new InvalidOperationException();
#endif
        }

        internal void AddRemovedBySource(TemporalGraphVertex source, IEnumerable<TemporalGraphVertex> oldTargets)
        {
            if (!_removedBySource.TryGetValue(source, out var targets))
            {
                targets = new List<TemporalGraphVertex>();
                _removedBySource.Add(source, targets);
            }

            targets.AddRange(oldTargets);
        }

        internal void AddRemovedBySource(TemporalGraphVertex source, TemporalGraphVertex oldTarget)
        {
            AddRemovedBySource(source, Enumerable.Repeat(oldTarget, 1));
        }

        internal void AddRemovedByTarget(TemporalGraphVertex target, IEnumerable<TemporalGraphVertex> oldSources)
        {
            if (!_removedByTarget.TryGetValue(target, out var sources))
            {
                sources = new List<TemporalGraphVertex>();
                _removedByTarget.Add(target, sources);
            }

            sources.AddRange(oldSources);
        }

        internal void AddRemovedByTarget(TemporalGraphVertex target, TemporalGraphVertex oldSource)
        {
            AddRemovedByTarget(target, Enumerable.Repeat(oldSource, 1));
        }
    }
}
