﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace TemporalSourceCover
{
#if DEBUG
    class TemporalGraphState
    {
        public readonly HashSet<TemporalGraphVertex> BlackVertices;
        public readonly HashSet<TemporalGraphVertex> RedVertices;
        public readonly Dictionary<TemporalGraphVertex, Time?> ReachedTimes;

        public readonly HashSet<TemporalGraphEdge> Edges;

        public readonly HashSet<TemporalGraphVertex> Normal;
        public readonly HashSet<TemporalGraphVertex> Sources;
        public readonly HashSet<TemporalGraphVertex> Sinks;
        public readonly HashSet<TemporalGraphVertex> Gateways;
        public readonly HashSet<TemporalGraphVertex> Isolated;

        public TemporalGraphState(TemporalGraph graph)
        {
            BlackVertices = new HashSet<TemporalGraphVertex>(graph.BlackVertices);
            RedVertices = new HashSet<TemporalGraphVertex>(graph.RedVertices);

            Edges = new HashSet<TemporalGraphEdge>(graph.Edges);

            Normal = new HashSet<TemporalGraphVertex>(graph.VerticesByType[VertexType.Regular]);
            Sources = new HashSet<TemporalGraphVertex>(graph.VerticesByType[VertexType.Source]);
            Sinks = new HashSet<TemporalGraphVertex>(graph.VerticesByType[VertexType.Sink]);
            Gateways = new HashSet<TemporalGraphVertex>(graph.VerticesByType[VertexType.Gateway]);
            Isolated = new HashSet<TemporalGraphVertex>(graph.VerticesByType[VertexType.Isolated]);

            ReachedTimes = graph.Vertices.Values.ToDictionary(v => v, v => v.ReachedAt);
        }

        private static bool SetsEqual<T>(HashSet<T> a, HashSet<T> b)
        {
            return a.Count == b.Count
                && a.All(t => b.Contains(t));
        }

        public static bool operator ==(TemporalGraphState a, TemporalGraphState b)
        {
            bool result = SetsEqual(a.BlackVertices, b.BlackVertices);
            result = result && SetsEqual(a.RedVertices, b.RedVertices);
            result = result && SetsEqual(a.Edges, b.Edges);
            result = result && SetsEqual(a.Normal, b.Normal);
            result = result && SetsEqual(a.Sources, b.Sources);
            result = result && SetsEqual(a.Sinks, b.Sinks);
            result = result && SetsEqual(a.Gateways, b.Gateways);
            result = result && SetsEqual(a.Isolated, b.Isolated);
            result = result && a.ReachedTimes.Count == b.ReachedTimes.Count;
            result = result && a.ReachedTimes.All(p => b.ReachedTimes[p.Key] == p.Value);

            return result;
        }

        public static bool operator !=(TemporalGraphState a, TemporalGraphState b) => !(a == b);

        public override bool Equals(object? obj)
        {
            return obj is TemporalGraphState state &&
                this == state;
        }
    }
#endif

    public sealed class TemporalGraphTransaction : IDisposable
    {
        private enum TransactionState
        {
            Active,
            Committed,
            RolledBack,
        }

        private readonly TemporalGraphTransaction? _outerTransaction;

        internal readonly List<TemporalGraphEdge> RemovedEdges = new();
        internal readonly List<TemporalGraphVertex> RemovedVertices = new();
        internal readonly Dictionary<TemporalGraphVertex, Time?> OldReachedTimes = new();

        private readonly TemporalGraph _graph;
        private TransactionState _state = TransactionState.Active;


#if DEBUG
        private readonly TemporalGraphState _initialState;
#endif

        public TemporalGraphTransaction(TemporalGraph graph)
        {
            _graph = graph;
            _outerTransaction = _graph.CurrentTransaction;

            _graph.CurrentTransaction = this;


#if DEBUG
            _initialState = new TemporalGraphState(_graph);
#endif
        }

        public void Dispose()
        {
            if (_state == TransactionState.Active)
                throw new InvalidOperationException();

            _graph.CurrentTransaction = _outerTransaction;
        }

        public void Commit()
        {
            if (_state != TransactionState.Active)
                throw new InvalidOperationException();

            _state = TransactionState.Committed;
        }

        public void Rollback()
        {
            if (_state != TransactionState.Active)
                throw new InvalidOperationException();

            using (_graph.DontUpdate())
            {
                foreach (var vertex in RemovedVertices)
                {
                    _graph.AddVertex(vertex);
                }

                foreach (var edge in RemovedEdges)
                {
                    _graph.AddEdge(edge);
                }
            }
            foreach (var (vertex, time) in OldReachedTimes)
            {
                vertex.ReachedAt = time;
            }


            _state = TransactionState.RolledBack;

#if DEBUG
            var curState = new TemporalGraphState(_graph);
            if (_initialState != curState)
                throw new InvalidOperationException();
#endif
        }
    }
}
