﻿using System;
using System.Collections.Generic;
using System.IO;

namespace TemporalSourceCover
{
    public static partial class CsvOutput
    {
        public static class Heuristics
        {
            public static string OutputPath { get; set; } = "output.csv";

            private static Scope? _scope;
            public static Scope CurrentScope => _scope ?? throw new NullReferenceException();

            private static List<Scope> _scopes = new();


            private static readonly DataReductionRule[] _dataReductionRules = Enum.GetValues<DataReductionRule>();
            private static readonly string _decimalFormat = "0.0#########";


            private static bool _initialized = false;

            private static void TryInit(Scope scope)
            {
                if (_initialized)
                    return;

                var headers = new List<string>
                {
                    "Ok?",
                    "",

                    "Name",
                    "Directed?",
                    "Mode",
                    "",
                };

                foreach (var heuristic in scope.Stats)
                {
                    headers.AddRange(new[]
                    {
                        $"{heuristic.Name} OK?",
                        $"{heuristic.Name} Time (in s)",
                        $"{heuristic.Name} Solution Size",
                    });
                }

                File.WriteAllLines(OutputPath, new[] { string.Join(";", headers) });

                _initialized = true;
            }

            private static void Write(Scope scope)
            {
                TryInit(scope);

                var data = new List<object>
                {
                    scope.Ok ? "Yes" : "No",
                    "",

                    // dataset info
                    scope.Name,
                    scope.Directed ? "Yes" : "No",
                    scope.Mode,
                    "",
                };

                foreach (var heuristic in scope.Stats)
                {
                    data.AddRange(new object[]
                    {
                        heuristic.Ok ? "Yes" : "No",
                        heuristic.Time.TotalSeconds.ToString(_decimalFormat),
                        heuristic.SolutionSize,
                    });
                }
                File.AppendAllLines(OutputPath, new[] { string.Join(";", data) });
            }

            public sealed class Scope : IDisposable
            {
                public bool Ok { get; set; } = true;

                public string Name { get; set; } = "";
                public bool Directed { get; set; } = true;
                public TimeMode Mode { get; set; } = TimeMode.Strict;

                public HeuristicStats Stats { get; } = new HeuristicStats();

                public class HeuristicStats : List<HeuristicStat>
                {
                }

                public class HeuristicStat
                {
                    public bool Ok { get; }
                    public string Name { get; }
                    public TimeSpan Time { get; }
                    public int SolutionSize { get; }

                    public HeuristicStat(string name)
                    {
                        Ok = false;
                        Name = name;
                    }

                    public HeuristicStat(string name, TimeSpan time, int solutionSize)
                    {
                        Ok = true;
                        Name = name;
                        Time = time;
                        SolutionSize = solutionSize;
                    }
                }


                public Scope()
                {
                    if (_scope is not null)
                        throw new InvalidOperationException();

                    _scope = this;
                }

                public void Dispose()
                {
                    _scopes.Add(this);
                    Write(this);

                    _scope = null;
                }
            }
        }
    }
}
