﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace TemporalSourceCover
{
    public static partial class CsvOutput
    {

        public static string OutputPath { get; set; } = "output.csv";

        private static Scope? _scope;
        public static Scope CurrentScope => _scope ?? throw new NullReferenceException();

        private static List<Scope> _scopes = new();
        
        
        private static readonly DataReductionRule[] _dataReductionRules = Enum.GetValues<DataReductionRule>();
        private static readonly string _decimalFormat = "0.0#########";

        public sealed class Scope : IDisposable
        {
            public bool Ok { get; set; } = true;

            public string Name { get; set; } = "";
            public bool Directed { get; set; } = true;
            public Interval Lifetime { get; set; }
            public TimeMode Mode { get; set; } = TimeMode.Strict;

            /// <summary>
            /// Time to read the input and create the graph model
            /// </summary>
            public TimeSpan? ReadInputTime { get; set; }

            public TimeSpan? DataReductionTime { get; set; }

            public GurobiStats InitialGurobiStats { get; } = new GurobiStats();
            public GurobiStats GurobiStatsAfterDataReduction { get; } = new GurobiStats();

            public SearchTreeStats SearchTreeAfterDataReduction { get; } = new SearchTreeStats();



            public GraphStats Initial { get; } = new GraphStats();
            public GraphStats AfterDataReduction { get; } = new GraphStats();


            public int PossibleSourcesAfterDataReduction { get; set; }

            public DataReductionStats DataReduction { get; } = new DataReductionStats();


            public class DataReductionRuleStats
            {
                public int TimesApplied { get; set; }
                public long RemovedVertices { get; set; }
                public long RemovedEdges { get; set; }
                public long VerticesAddedToSolution { get; set; }
                public long SupersetIgnored { get; set; }
                public long SubsetIgnored { get; set; }
                public TimeSpan Time { get; set; }
            }

            public class DataReductionStats : Dictionary<DataReductionRule, DataReductionRuleStats>
            {
                public DataReductionStats()
                {
                    foreach (var reductionRule in Enum.GetValues<DataReductionRule>())
                    {
                        Add(reductionRule, new DataReductionRuleStats());
                    }
                }
            }

            public class SearchTreeStats
            {
                public bool UseReductionRulesTemporal;
                public bool UseReductionRulesReachability;

                public int SolutionSize;
                public long Calls;
                public TimeSpan? RunningTime;
            }

            public class GurobiStats
            {
                public int SolutionSize;
                public int ReachabilityGraphEdges;
                public TimeSpan? ModelBuildTime;
                public TimeSpan? RunningTime;
            }

            public class GraphStats
            {
                public int VerticesInSolution;

                public int VertexCount;
                public int EdgeCount;

                public int BlackVertices;
                public int RedVertices;

                public Dictionary<VertexType, int> CountByVertexType { get; } = Enum.GetValues<VertexType>().ToDictionary(v => v, _ => 0);
            }


            public Scope()
            {
                if (_scope is not null)
                    throw new InvalidOperationException();

                _scope = this;
            }

            public void Dispose()
            {
                _scopes.Add(this);
                Write(this);

                _scope = null;
            }
        }

        private static bool _initialized = false;

        private static void TryInit()
        {
            if (_initialized)
                return;

            var headers = new List<string>
            {
                "OK?",
                "",

                // dataset info
                "Name",
                "Directed",
                "Lifetime",
                "Mode",
                "Time to initialize graph (in s)",
                "",

                // initial graph stats
                "#Vertices",
                "#Edges",
                "#Normal Vertices",
                "#Source Vertices",
                "#Sink Vertices",
                "#Gateway Vertices",
                "",

                // initial gurobi stats
                "Gurobi Solution Size",
                "Time to build Gurobi model (in s)",
                "Gurobi running time (in s)",
                "Reachability Graph #Edges",
                "",

                // initial data reduction stats
                "Time to run initial data reduction (in s)",
            };

            foreach (var rule in _dataReductionRules)
            {
                headers.AddRange(new[]
                {
                    $"({rule}) Times applied",
                    $"({rule}) Total Time",
                    $"({rule}) Removed vertices",
                    $"({rule}) Removed edges",
                    $"({rule}) Vertices added to solution",
                    $"({rule}) Vertices ignored for solution",
                });
            }

            headers.AddRange(new[]
            {
                "",

                // graph stats after initial data reduction
                "#Vertices in Solution (after data reduction)",
                "#Sources to checked (after data reduction)",
                "#Vertices (after data reduction)",
                "#Edges (after data reduction)",
                "#Normal Vertices (after data reduction)",
                "#Source Vertices (after data reduction)",
                "#Sink Vertices (after data reduction)",
                "#Gateway Vertices (after data reduction)",
                "#Black Vertices (after data reduction)",
                "#Red Vertices (after data reduction)",
                "",

                // gurobi stats after initial data reduction
                "Gurobi Solution Size (after data reduction)",
                "Time to build Gurobi model (in s) (after data reduction)",
                "Gurobi running time (in s) (after data reduction)",
                "Reachability Graph #Edges (after data reduction)",
                "",

                // search tree stats
                "Temporal Rules?",
                "Reachability Rules?",
                "Search Tree Calls",
                "Search Tree Solution Size",
                "Search Tree Time (in s)",
            });
            File.WriteAllLines(OutputPath, new[] { string.Join(";", headers) });

            _initialized = true;
        }

        private static void Write(Scope scope)
        {
            TryInit();

            var data = new List<object>
            {
                scope.Ok ? "Yes" : "No",
                "",

                // dataset info
                scope.Name,
                scope.Directed ? "Yes" : "No",
                (scope.Lifetime.Max.Value - scope.Lifetime.Min.Value),
                scope.Mode,
                scope.ReadInputTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
                "",

                // initial graph stats
                scope.Initial.VertexCount,
                scope.Initial.EdgeCount,
                scope.Initial.CountByVertexType[VertexType.Regular],
                scope.Initial.CountByVertexType[VertexType.Source],
                scope.Initial.CountByVertexType[VertexType.Sink],
                scope.Initial.CountByVertexType[VertexType.Gateway],
                "",

                // initial gurobi stats
                scope.InitialGurobiStats.ModelBuildTime is object ? scope.InitialGurobiStats.SolutionSize.ToString() : "-",
                scope.InitialGurobiStats.ModelBuildTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
                scope.InitialGurobiStats.RunningTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
                scope.InitialGurobiStats.ReachabilityGraphEdges.ToString() ?? "ABORTED",
                "",

                // initial data reduction stats
                scope.DataReductionTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
            };

            foreach (var rule in _dataReductionRules)
            {
                var reduction = scope.DataReduction[rule];
                data.AddRange(new object[]
                {
                    reduction.TimesApplied,
                    reduction.Time.TotalSeconds.ToString(_decimalFormat),
                    reduction.RemovedVertices,
                    reduction.RemovedEdges,
                    reduction.VerticesAddedToSolution,
                    reduction.SubsetIgnored,
                });
            }

            data.AddRange(new object[]
            {
                "",

                // graph stats after initial data reduction
                scope.AfterDataReduction.VerticesInSolution,
                scope.PossibleSourcesAfterDataReduction,
                scope.AfterDataReduction.VertexCount,
                scope.AfterDataReduction.EdgeCount,
                scope.AfterDataReduction.CountByVertexType[VertexType.Regular],
                scope.AfterDataReduction.CountByVertexType[VertexType.Source],
                scope.AfterDataReduction.CountByVertexType[VertexType.Sink],
                scope.AfterDataReduction.CountByVertexType[VertexType.Gateway],
                scope.AfterDataReduction.BlackVertices,
                scope.AfterDataReduction.RedVertices,
                "",

                // gurobi stats after initial data reduction
                scope.GurobiStatsAfterDataReduction.ModelBuildTime is object ? scope.GurobiStatsAfterDataReduction.SolutionSize.ToString(_decimalFormat) : "-",
                scope.GurobiStatsAfterDataReduction.ModelBuildTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
                scope.GurobiStatsAfterDataReduction.RunningTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
                scope.GurobiStatsAfterDataReduction.ReachabilityGraphEdges.ToString() ?? "ABORTED",
                "",
                
                // search tree stats
                scope.SearchTreeAfterDataReduction.UseReductionRulesTemporal ? "Yes" : "No", 
                scope.SearchTreeAfterDataReduction.UseReductionRulesReachability ? "Yes" : "No",
                scope.SearchTreeAfterDataReduction.Calls,
                scope.SearchTreeAfterDataReduction.SolutionSize,
                scope.SearchTreeAfterDataReduction.RunningTime?.TotalSeconds.ToString(_decimalFormat) ?? "ABORTED",
                "Search Tree Time (in s)",
            });
            File.AppendAllLines(OutputPath, new[] { string.Join(";", data) });
        }
    }
}
