﻿using System;
using System.Collections.Generic;
using System.Linq;
using TemporalSourceCover.Utility;

namespace TemporalSourceCover
{
    public static partial class DataReductionRules
    {
        private delegate bool ReachabilityRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result);

        private static readonly ReachabilityRule[] _reachabilityRules =
        {
            RedReachabilityRule,

            IgnoreGatewaysAndSinksAsSourceRule,
            RemoveTargetsCoveredBySuperset,
            ReachabilitySubsetRule,

            SingleSourceRule,
        };

        private static bool ApplyReachabilityReductions(TemporalSourceCoverInstance instance, Solution solution, List<DataReductionResult> reductions)
        {
            if (instance.Reachability is null)
            {
                instance.Reachability = instance.Graph.GetFullReachability();
            }


            bool didReduce;
            do
            {
                didReduce = false;
                foreach (var rule in _reachabilityRules)
                {
                    Timeout.ThrowOnCancel();

                    using var metrics = Metrics.BeginMeasurement(null, log: false);

                    var applied = rule.Invoke(instance, solution, out var reduction);

                    reduction.Time = metrics.Time;
                    reduction.WasApplied = applied;
                    reductions.Add(reduction);

                    didReduce |= applied;
                }
            } while (didReduce);

            return didReduce;
        }

        private static bool IgnoreGatewaysAndSinksAsSourceRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.IgnoreGatewaysAndSinksAsSource);
            bool applied = false;

            var gatewaysAndSinks = instance.Graph.VerticesByType[VertexType.Gateway]
                .Concat(instance.Graph.VerticesByType[VertexType.Sink]);

            foreach (var vertex in gatewaysAndSinks)
            {
                if (!instance.Reachability!.BySource.TryGetValue(vertex, out var targets))
                    continue;

                applied |= instance.Reachability.RemoveSource(vertex);
                result.IgnoredVertices++;
            }

            return applied;
        }


        private static bool SingleSourceRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.SingleSource);
            bool applied = false;

            var targetPairs = instance.Reachability!.ByTarget.ToList();
            foreach (var (target, sources) in targetPairs)
            {
                if (sources.Count > 1)
                    continue;

                var singleSource = sources.First();
                // source was single source for earlier vertex already
                if (!instance.Reachability.BySource.TryGetValue(singleSource, out var targets))
                    continue;

                instance.Graph.RemoveSourceAndPaths(singleSource, out _, out _);

                foreach (var reachedTarget in targets)
                {
                    instance.Reachability.RemoveTarget(reachedTarget);
                }
                instance.Reachability.RemoveSource(singleSource);

                solution.Add(singleSource);
                result.SolutionVertices++;
                instance.K--;

                applied = true;

                if (instance.K < 0)
                    break;
            }

            return applied;
        }

        private static bool ReachabilitySubsetRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.ReachabilitySubset);
            bool applied = false;

            // sort sources ascending by number of reachable vertices
            var sourcePairs = instance.Reachability!.BySource.ToList();
            sourcePairs.Sort((a, b) => a.Value.Count.CompareTo(b.Value.Count));

            for (int i = 0; i < sourcePairs.Count; i++)
            {
                var (source, targets) = sourcePairs[i];

                bool isSubset = false;
                for (int j = i + 1; j < sourcePairs.Count; j++)
                {
                    var (otherSource, otherTargets) = sourcePairs[j];

                    if (targets.All(t => t.IsRed || otherTargets.Contains(t)))
                    {
                        isSubset = true;
                        break;
                    }
                }

                // remove source of subset from reachability
                if (isSubset)
                {
                    result.IgnoredVertices++;
                    instance.Reachability.RemoveSource(source);

                    applied = true;
                }
            }

            return applied;
        }

        private static bool RemoveTargetsCoveredBySuperset(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.Reachability_TargetsCoveredBySuperset);
            bool applied = false;

            var coveredTargets = new HashSet<TemporalGraphVertex>();

            // sort asecending by number of sources to ensure that covering superset are at a higher index
            var pairsByTarget = instance.Reachability!.ByTarget.ToList();
            pairsByTarget.Sort((a, b) => a.Value.Count.CompareTo(b.Value.Count));

            // ignore targets that are covered by supersets
            for (int i = 0; i < pairsByTarget.Count; i++)
            {
                var (target, sources) = pairsByTarget[i];

                // skip already reached or covered vertices
                if (target.IsRed || coveredTargets.Contains(target))
                    continue;

                // check for super-set of sources => super-set is covered anyways, ignore target then
                for (int j = i + 1; j < pairsByTarget.Count; j++)
                {
                    var (otherTarget, otherSources) = pairsByTarget[j];
                    if (!sources.All(src => otherSources.Contains(src)))
                        continue;

                    coveredTargets.Add(otherTarget);
                    // only remove from targets and not for reachability of the sources
                    instance.Reachability.RemoveOnlyTarget(otherTarget);

                    applied = true;

                    result.IgnoredVertices++;
                }
            }

            return applied;
        }

        /// <summary>
        /// This rule removes all red vertices from the to-reach information for all vertices.
        /// This effectively applied the red-reachability rule as well as additionally removes red vertices from the sets of reachable vertices.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <param name="solution"></param>
        /// <param name="reachability"></param>
        /// <param name="result"></param>
        /// <returns></returns>
        private static bool RedReachabilityRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.RedReachability);
            bool applied = false;

            var reachability = instance.Reachability!;

            // remove red vertices as targets
            foreach (var target in instance.Graph.RedVertices)
            {
                applied |= reachability.RemoveTarget(target);
            }

            // remove sources that only reach red vertices
            foreach (var (source, targets) in reachability.BySource.ToList())
            {
                if (targets.All(v => v.IsRed))
                {
                    reachability.RemoveSource(source);
                }
            }

            return applied;
        }
    }
}
