﻿using Assertions;
using System.Collections.Generic;
using System.Linq;
using TemporalSourceCover.Utility;

namespace TemporalSourceCover
{
    public static partial class DataReductionRules
    {
        private delegate bool TemporalRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result);

        private static readonly TemporalRule[] _temporalRules =
        {
            BlackSourceRule,
            RedSinkRule,
            DegreeZeroRules,
            EdgeToReachedSubgraphRule,
            //RedundantEdgeRule,
            SingleNeighborRule,
        };

        private static bool ApplyTemporalSourceCoverDataReductions(TemporalSourceCoverInstance instance, Solution solution, List<DataReductionResult> reductions)
        {
            bool didReduce;
            do
            {
                didReduce = false;
                foreach (var rule in _temporalRules)
                {
                    Timeout.ThrowOnCancel();

                    using var metrics = Metrics.BeginMeasurement(null, log: false);

                    var applied = rule.Invoke(instance, solution, out var reduction);

                    reduction.Time = metrics.Time;
                    reduction.WasApplied = applied;
                    reductions.Add(reduction);

                    didReduce |= applied;

                    if (instance.K < 0)
                        break;
                }
            } while (didReduce);
            return didReduce;
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <param name="solution"></param>
        /// <param name="result"></param>
        /// <returns></returns>
        private static bool SingleNeighborRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.SingleNeighbor);
            var vertices = instance.Graph.Vertices.Values.ToList();
            bool applied = false;

            foreach (var vertex in vertices)
            {
                Timeout.ThrowOnCancel();

                if (vertex.InNeighbors.Count != 1)
                    continue;
                if (vertex.OutNeighbors.Count != 1)
                    continue;
                if (vertex.InNeighbors.Keys.First() != vertex.OutNeighbors.Keys.First())
                    continue;


                foreach (var outEdge in vertex.OutEdges.ToList())
                {
                    instance.Graph.RemoveEdge(outEdge);
                    result.RemovedEdges++;

                    applied = true;
                }
            }

            return applied;
        }

        private static bool RedundantEdgeRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.RedundantEdge);
            bool applied = false;

            foreach (var vertex in instance.Graph.Vertices.Values)
            {
                var edgeTimes = new Dictionary<(Time? inTime, Time? outTime, TemporalGraphVertex to), TemporalGraphEdge>();
                var removeEdges = new HashSet<TemporalGraphEdge>();

                foreach (var edge in vertex.OutEdges)
                {
                    var latestEarlierInEdge = edge.From.InEdges
                        .TakeWhile(inEdge => Time.Proceeds(inEdge.Time, edge.Time))
                        .LastOrDefault();

                    var earliestLaterOutEdge = edge.To.OutEdges
                        .SkipWhile(outEdge => !Time.Proceeds(edge.Time, outEdge.Time))
                        .FirstOrDefault();

                    var key = (latestEarlierInEdge?.Time, earliestLaterOutEdge?.Time, edge.To);
                    if (!edgeTimes.TryGetValue(key, out var keyEdge))
                    {
                        edgeTimes.Add(key, edge);
                    }
                    else if (keyEdge.Time > edge.Time)
                    {
                        removeEdges.Add(keyEdge);
                        edgeTimes[key] = edge;
                    }
                    else
                    {
                        removeEdges.Add(edge);
                    }
                }

                foreach (var removeEdge in removeEdges)
                {
                    applied = true;
                    instance.Graph.RemoveEdge(removeEdge);

                    result.RemovedEdges++;
                }
            }

            return applied;
        }

        /// <summary>
        /// Removes all in-edges of a red vertex that exists at a timestep greater or equal to the timestep the vertex was reached.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <param name="solution"></param>
        /// <param name="result"></param>
        /// <returns></returns>
        private static bool EdgeToReachedSubgraphRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.EdgeToReachedSubgraph);

            var redVertices = instance.Graph.RedVertices.Where(v => v.InEdges.Count > 0);

            var applied = false;

            foreach (var redVertex in redVertices)
            {
                Timeout.ThrowOnCancel();

                var removeEdges = redVertex.InEdges.SkipWhile(e => e.Time < redVertex.ReachedAt).ToList();
                foreach (var removeEdge in removeEdges)
                {
                    Timeout.ThrowOnCancel();

                    instance.Graph.RemoveEdge(removeEdge);
                    result.RemovedEdges++;

                    applied = true;
                }
            }


            return applied;
        }

        /// <summary>
        /// Isolated vertices can simply be removed from the graph.
        /// If they are black they will be put into the solution.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <param name="solution"></param>
        /// <returns></returns>
        internal static bool DegreeZeroRules(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            result = new DataReductionResult(DataReductionRule.Isolated);
            bool applied = false;

            var isolatedVertices = instance.Graph.VerticesByType[VertexType.Isolated].ToList();
            foreach (var vertex in isolatedVertices)
            {
                Timeout.ThrowOnCancel();

                if (!vertex.IsRed)
                {
                    solution.Add(vertex);
                    result.SolutionVertices++;
                    instance.K--;
                }

                Assert.Zero(instance.Graph.RemoveVertex(vertex));
                result.RemovedVertices++;

                applied = true;

                if (instance.K < 0)
                    break;
            }

            return applied;
        }

        /// <summary>
        /// Red sinks can simply be removed from the graph.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <param name="solution"></param>
        /// <returns></returns>
        internal static bool RedSinkRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            bool applied = false;
            result = new DataReductionResult(DataReductionRule.RedSink);

            var redSinks = instance.Graph.VerticesByType[VertexType.Sink].Where(s => s.IsRed).ToList();
            foreach (var vertex in redSinks)
            {
                Timeout.ThrowOnCancel();

                result.RemovedEdges += instance.Graph.RemoveVertex(vertex);
                result.RemovedVertices++;

                applied = true;
            }

            return applied;
        }

        /// <summary>
        /// Removes black sources and all edges that can be reached from the source.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <param name="solution"></param>
        /// <returns></returns>
        internal static bool BlackSourceRule(TemporalSourceCoverInstance instance, Solution solution, out DataReductionResult result)
        {
            bool applied = false;
            result = new DataReductionResult(DataReductionRule.BlackSource);

            var blackSources = instance.Graph.VerticesByType[VertexType.Source].Where(s => !s.IsRed).ToList();
            foreach (var vertex in blackSources)
            {
                Timeout.ThrowOnCancel();

                solution.Add(vertex);
                result.SolutionVertices++;
                instance.K--;

                instance.Graph.RemoveSourceAndPaths(vertex, out var reachedVertices, out var removedEdges);
                result.RemovedEdges += removedEdges.Count;
                result.RemovedVertices++;

                applied = true;

                if (instance.K < 0)
                    break;
            }

            return applied;
        }
    }
}
