﻿#if VERIFY
using Gurobi;
using System.Diagnostics;
using System.Threading.Tasks;
#endif
using System;
using System.Collections.Generic;
using System.Linq;
using TemporalSourceCover.Utility;

namespace TemporalSourceCover
{
    public static class TemporalSourceCover
    {
        public static bool UseSearchTree = true;
        public static bool UseDataReductionRules = true;

        #if VERIFY
        /// <summary>
        /// Uses Gurobi to solve the temporal source cover problem using directed dominating set.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <returns></returns>
        public static Solution FindGurobiSolution(TemporalSourceCoverInstance instance, CsvOutput.Scope.GurobiStats stats, double epsilon = 1e-6, TimeSpan? timeout = default)
        {
            var temporalGraph = instance.Graph;

            using var env = new GRBEnv(true)
            {
                LogToConsole = 0
            };

            env.Start();

            using var model = new GRBModel(env);
            if (timeout is TimeSpan t)
            {
                model.Parameters.TimeLimit = t.TotalSeconds;
            }

            Dictionary<TemporalGraphVertex, GRBVar> variables = null!;


            model.Parameters.OptimalityTol = epsilon;
            model.Parameters.IntFeasTol = epsilon;

            var possibleSources = instance.PossibleSources.ToList();

            var modelBuildTime = TimeSpan.Zero;

            var modelStopwatch = Stopwatch.StartNew();
            variables = temporalGraph.Vertices.Values.ToDictionary(
                v => v,
                v => model.AddVar(0, 1, 0, GRB.BINARY, $"vertex{v.Id}"));
            var constraints = temporalGraph.Vertices.Values.ToDictionary(
                v => v,
                v => new GRBLinExpr());
            var objective = new GRBLinExpr();

            var objectiveLock = new object();
            var constraintLock = new object();
            var reachabilityTimeLock = new object();
            var modelTimeLock = new object();

            // get reachability by vertex and add it as source variable to the reached vertices
            Parallel.ForEach(possibleSources, v =>
            {
                Timeout.ThrowOnCancel();

                var variable = variables[v];
                lock (objectiveLock)
                {
                    objective += variable;
                }

                IEnumerable<TemporalGraphVertex> reachableVertices;
                if (instance.Reachability is null)
                {
                    reachableVertices = temporalGraph.GetReachableVertices(v);
                }
                else
                {
                    reachableVertices = instance.Reachability.BySource.TryGetValue(v, out var targets) ? targets : Enumerable.Empty<TemporalGraphVertex>();
                }


                foreach (var reached in reachableVertices)
                {
                    lock (constraintLock)
                    {
                        stats.ReachabilityGraphEdges++;
                        constraints[reached] += variable;
                    }
                }
            });

            model.SetObjective(objective, GRB.MINIMIZE);

            // add the built contraints
            foreach (var (v, contraint) in constraints)
            {
                Timeout.ThrowOnCancel();

                double target = v.IsRed ? 0.0 : 1.0;
                model.AddConstr(contraint >= target, $"contraint_for_vertex{v.Id}");
            }
            modelStopwatch.Stop();
            modelBuildTime = modelStopwatch.Elapsed;

            stats.ModelBuildTime = modelBuildTime;
            Console.WriteLine($"Building Gurobi Model: {modelBuildTime.TotalSeconds}s (inclduing reachability)");


            using (var metrics = Metrics.BeginMeasurement(null, log: false))
            {
                model.Optimize();

                if (model.Status == GRB.Status.TIME_LIMIT)
                    throw new TaskCanceledException();

                if (model.Status != GRB.Status.OPTIMAL)
                    throw new InvalidOperationException();

                stats.RunningTime = metrics.Time;
                Console.WriteLine($"Solving ILP: {metrics.Time.TotalSeconds} s");
            }


            var dominatingSet = variables.Where(pair => pair.Value.X > 0).ToList();

            var solution = new Solution(dominatingSet.Select(pair => pair.Key));
            stats.SolutionSize = solution.Count;

            return solution;
        }
#endif

        public static Solution FindSolutionSearchTree(TemporalSourceCoverInstance rootInstance, out long calls)
        {
            calls = 0;

            var graph = rootInstance.Graph;
            var reachability = rootInstance.Reachability;
            var forbiddenVertices = rootInstance.ForbiddenVertices;

            var minK = graph.VerticesByType[VertexType.Source].Count(v => !v.IsRed)
                + graph.VerticesByType[VertexType.Isolated].Count(v => !v.IsRed);

            for (int k = minK; k <= graph.BlackVertices.Count; k++)
            {
                Timeout.ThrowOnCancel();

                var instance = new TemporalSourceCoverInstance(graph, k, new HashSet<TemporalGraphVertex>(forbiddenVertices))
                {
                    Reachability = reachability?.Clone(),
                };

                Console.WriteLine($"Search Tree: Trying to find solution of size {k}");

                var solution = TryFindSolutionSearchTree(instance, 0, ref calls, new HashSet<TemporalGraphVertex>());
                if (solution is not null)
                {
                    Console.WriteLine($"Search Tree Recursive calls: {calls}");
                    return solution;
                }
            }

            throw new InvalidOperationException();
        }

        private static Solution? TryFindSolutionSearchTree(TemporalSourceCoverInstance instance, int depth, ref long calls, HashSet<TemporalGraphVertex> triedSources)
        {
            Timeout.ThrowOnCancel();

            calls++;

            if (instance.K < 0)
                return null;


            bool edgeToUnreached = instance.Graph.Edges.Except(instance.Graph.RedVertices.SelectMany(r => r.InEdges.Concat(r.OutEdges))).Any();
            if (!edgeToUnreached)
            {
                if (instance.Graph.BlackVertices.Count > instance.K)
                    return null;

                return new Solution(instance.Graph.BlackVertices);
            }

            if (instance.K == 0)
                return null;


            using var transaction = new TemporalSourceCoverTransaction(instance);

            var reductionSolution = new Solution();
            if (UseDataReductionRules)
            {
                // apply data reduction, if more than k vertices added to solution => return false
                reductionSolution = DataReductionRules.Apply(instance, out _);
                if (instance.K < 0)
                {
                    transaction.Rollback();
                    return null;
                }

                // if no black vertices left after data reduction, return true
                if (instance.Graph.BlackVertices.Count == 0)
                {
                    transaction.Commit();
                    return reductionSolution;
                }
            }

            var possibleSources = instance.PossibleSources.Except(triedSources).ToList();
            //possible heuristic: sort possible sources by earliest out-edge (with isolated vertices being in front)
            //possibleSources.Sort((a, b) =>
            //{
            //    if (a.Type == VertexType.Isolated)
            //        return -1;

            //    if (b.Type == VertexType.Isolated)
            //        return 1;

            //    var aFirstEdge = a.OutEdges.Min ?? throw new InvalidOperationException();
            //    var bFirstEdge = b.OutEdges.Min ?? throw new InvalidOperationException();
            //    return aFirstEdge.Time.CompareTo(bFirstEdge.Time);
            //});

#if DEBUG
            Console.WriteLine(new string(' ', depth) + $"{possibleSources.Count} possible sources");
#endif

            var innerTriedSources = new HashSet<TemporalGraphVertex>(triedSources);

            var oldK = instance.K;
            var k = instance.K - reductionSolution.Count - 1;
            foreach (var vertex in possibleSources)
            {
                instance.K = k;

                using var innerTransaction = new TemporalSourceCoverTransaction(instance);

                instance.Graph.RemoveSourceAndPaths(vertex, out _, out _);
                instance.Reachability?.RemoveSource(vertex);

                innerTriedSources.Add(vertex);

                if (TryFindSolutionSearchTree(instance, depth + 1, ref calls, innerTriedSources) is Solution solution)
                {
                    solution.AddRange(reductionSolution);
                    solution.Add(vertex);

                    innerTransaction.Commit();
                    transaction.Commit();
                    return solution;
                }
                else
                {
                    innerTransaction.Rollback();
                }
            }

            transaction.Rollback();
            return null;
        }
    }
}
