﻿using Priority_Queue;
using System.Collections.Generic;
using System.Linq;
using TemporalSourceCover.Utility;

namespace TemporalSourceCover
{
    public static class TemporalSourceCoverHeuristicSolutions
    {
        /// <summary>
        /// Takes the vertex with the earliest out-edge if it reaches a black vertex.
        /// If no edges are left it takes all remaining black vertices.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <returns></returns>
        public static Solution EarliestOutEdge(TemporalGraph temporalGraph)
        {
            var solution = new Solution();

            var reached = temporalGraph.Vertices.Values.ToDictionary(v => v, _ => false);

            while (temporalGraph.BlackVertices.Count > 0)
            {
                Timeout.ThrowOnCancel();

                var edge = temporalGraph.Edges.FirstOrDefault();

                var source = edge?.From ?? temporalGraph.BlackVertices.First();
                temporalGraph.RemoveSourceAndPaths(source, out var reachedVertices, out _);
                if (reachedVertices.Any(v => !reached[v]))
                {
                    solution.Add(source);
                }

                foreach (var v in reachedVertices)
                {
                    reached[v] = true;
                }
            }

            return solution;
        }

        private class TemporalGraphVertexNode : FastPriorityQueueNode
        {
            public TemporalGraphVertex Vertex { get; }

            public TemporalGraphVertexNode(TemporalGraphVertex vertex)
            {
                Vertex = vertex;
            }
        }

        /// <summary>
        /// Takes the vertex with the highest number of black out-neighbors incl. itself
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <returns></returns>
        public static Solution LargestUnreachedOutNeighborhood(TemporalGraph temporalGraph)
        {
            var solution = new Solution();

            // init prio queue
            var queue = new FastPriorityQueue<TemporalGraphVertexNode>(temporalGraph.Vertices.Count);
            var nodes = new Dictionary<TemporalGraphVertex, TemporalGraphVertexNode>(temporalGraph.Vertices.Count);
            foreach (var vertex in temporalGraph.Vertices.Values)
            {
                Timeout.ThrowOnCancel();

                var node = new TemporalGraphVertexNode(vertex);
                var prio = int.MaxValue - (1 + vertex.OutNeighbors.Count);

                queue.Enqueue(node, prio);
                nodes.Add(vertex, node);
            }

            while (temporalGraph.BlackVertices.Count > 0)
            {
                Timeout.ThrowOnCancel();

                var node = queue.Dequeue();
                var source = node.Vertex;
                nodes.Remove(source);

                temporalGraph.RemoveSourceAndPaths(source, out var reachedVertices, out _);
                solution.Add(source);

                foreach (var reachedVertex in reachedVertices)
                {
                    if (reachedVertex == source)
                        continue;

                    if (nodes.TryGetValue(reachedVertex, out var reachedNode) && nodes.Remove(reachedVertex))
                    {
                        queue.Remove(reachedNode);
                    }
                }

                var update = new HashSet<TemporalGraphVertex>(reachedVertices.SelectMany(v => v.InNeighbors.Keys));
                foreach (var updateVertex in update)
                {
                    Timeout.ThrowOnCancel();

                    if (updateVertex.IsRed)
                        continue;

                    var updateNode = nodes[updateVertex];
                    var prio = int.MaxValue -
                        ((updateVertex.IsRed ? 0 : 1) + updateVertex.InNeighbors.Count(nodes => !nodes.Key.IsRed));

                    queue.UpdatePriority(updateNode, prio);
                }
            }

            return solution;
        }

        /// <summary>
        /// Takes the vertex that has the highest number of neighbors, if it reaches a black vertex.
        /// </summary>
        /// <param name="temporalGraph"></param>
        /// <returns></returns>
        public static Solution LargestOutNeighborhood(TemporalGraph temporalGraph)
        {
            var solution = new Solution();

            // init prio queue
            var queue = new FastPriorityQueue<TemporalGraphVertexNode>(temporalGraph.Vertices.Count);
            var nodes = new Dictionary<TemporalGraphVertex, TemporalGraphVertexNode>(temporalGraph.Vertices.Count);
            foreach (var vertex in temporalGraph.Vertices.Values)
            {
                Timeout.ThrowOnCancel();

                var node = new TemporalGraphVertexNode(vertex);
                var prio = int.MaxValue - vertex.OutNeighbors.Count;

                queue.Enqueue(node, prio);
                nodes.Add(vertex, node);
            }

            var reached = temporalGraph.Vertices.Values.ToDictionary(v => v, _ => false);

            while (temporalGraph.BlackVertices.Count > 0)
            {
                Timeout.ThrowOnCancel();

                var node = queue.Dequeue();
                var source = node.Vertex;

                temporalGraph.RemoveSourceAndPaths(source, out var reachedVertices, out var removedEdges);
                if (reachedVertices.Any(v => !reached[v]))
                {
                    solution.Add(source);
                }
                foreach (var v in reachedVertices)
                {
                    reached[v] = true;
                }

                var update = new HashSet<TemporalGraphVertex>(source.InNeighbors.Keys);
                foreach (var updateVertex in update)
                {
                    Timeout.ThrowOnCancel();

                    var updateNode = nodes[updateVertex];
                    var prio = int.MaxValue - updateVertex.OutNeighbors.Count;

                    queue.UpdatePriority(updateNode, prio);
                }
            }

            return solution;
        }
    }
}
